from __future__ import annotations

import copy
from enum import StrEnum
from typing import Any, Dict, List

import httpx
from pydantic import BaseModel, Field

from .resource_abc import CamelAlias, Resource, register_resource


class CollectionType(StrEnum):
    SINGLE = "Single"
    ARRAY = "Array"


class LifeTime(StrEnum):
    PERPETUAL = "Perpetual"
    TIMEVARIANT = "TimeVariant"


class FieldType(StrEnum):
    STRING = "String"
    BOOLEAN = "Boolean"
    DATE_TIME = "DateTime"
    DECIMAL = "Decimal"


class FieldDefinition(CamelAlias, BaseModel):
    name: str
    lifetime: LifeTime
    type: FieldType
    collection_type: CollectionType = CollectionType.SINGLE
    required: bool
    description: str = ""


# These are optional in the API create and will be given default values. When read is called
# they will not be returned if they have the default value


DEFAULT_FIELD = {"collectionType": "Single", "description": ""}


@register_resource()
class EntityTypeResource(CamelAlias, BaseModel, Resource):
    id: str = Field(exclude=True)
    entity_type_name: str
    display_name: str
    description: str
    field_schema: List[FieldDefinition]

    def read(self, client, old_state) -> Dict[str, Any]:
        entity_type = old_state.entitytype
        return client.request("get", f"/api/api/customentities/entitytypes/{entity_type}").json()

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        res = client.request("POST", "/api/api/customentities/entitytypes", json=desired).json()
        return {"entitytype": res["entityType"]}

    def update(self, client: httpx.Client, old_state):
        remote = self.read(client, old_state)
        # enrich remote fields with the default values if not present
        remote["fieldSchema"] = [rem | DEFAULT_FIELD for rem in remote["fieldSchema"]]
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        effective = remote | copy.deepcopy(desired)
        for i in range(0, len(self.field_schema)):
            if i < len(remote["fieldSchema"]):
                eff_field = remote["fieldSchema"][i] | desired["fieldSchema"][i]
                effective["fieldSchema"][i] = eff_field
        if effective == remote:
            return None
        res = client.request(
            "PUT", f"/api/api/customentities/entitytypes/{old_state.entitytype}", json=desired
        ).json()
        return {"entitytype": res["entityType"]}

    @staticmethod
    def delete(client, old_state):
        raise RuntimeError("Cannot delete a custom entity definition")

    def deps(self):
        return []



from .deploy import Deployment as Deployment
from .deploy import deploy as deploy
from .deploy import load_vars as load_vars
from .http_client import create_client as create_client
from .log import list_deployments as list_deployments
from .log import setup as setup



from fbnconfig import Deployment, identity


def configure(env):
    admin = identity.IdentityRoleRef(id="admin", name="instrument-administrator")
    operator = identity.IdentityRoleResource(
        id="role1", description="scheduler-operators", name="robTest_role2"
    )
    user = identity.UserResource(
        id="user_jane3",
        first_name="robTestJane",
        last_name="robTestSmith",
        email_address="robtest_jane@robert.byrnemail.org",
        login="robtest_jane3@robtest.com",
        type=identity.UserType.SERVICE,
    )
    assignment1 = identity.RoleAssignment(id="jane3_operator", user=user, role=operator)
    assignment2 = identity.RoleAssignment(id="jane3_admin", user=user, role=admin)
    return Deployment("identity", [assignment1, assignment2, user, operator])


if __name__ == "__main__":
    import os

    import click

    import fbnconfig

    @click.command()
    @click.argument("lusid_url", envvar="LUSID_ENV", type=str)
    @click.option("-v", "--vars_file", type=click.File("r"))
    def cli(lusid_url, vars_file):
        host_vars = fbnconfig.load_vars(vars_file)
        d = configure(host_vars)
        fbnconfig.deploy(d, lusid_url, os.environ["FBN_ACCESS_TOKEN"])

    cli()



import json
from types import SimpleNamespace

import httpx
import pytest

from fbnconfig import identity

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.raise_for_status()


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeIdentityRoleRef:
    # This should use the same method of creating a client as the host
    # To be refactored
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_when_exists(self, respx_mock):
        # given 2 roles in the system where role2 matches the sut
        respx_mock.get("/identity/api/roles").mock(
            return_value=httpx.Response(
                200,
                json=[
                    {
                        "id": "11111111",
                        "roleId": {"scope": "default", "code": "role1"},
                        "name": "role1",
                        "description": "role one",
                    },
                    {
                        "id": "22222222",
                        "roleId": {"scope": "default", "code": "role2"},
                        "name": "role2",
                        "description": "role two",
                    },
                ],
            )
        )
        client = self.client
        # when we attach
        sut = identity.IdentityRoleRef(id="xyz", name="role2")
        sut.attach(client)
        # then the roleId property is populated from the response
        assert sut.role_id == "22222222"

    def test_attach_when_not_exists(self, respx_mock):
        # given 2 roles in the system where neither matches the sut
        respx_mock.get("/identity/api/roles").mock(
            return_value=httpx.Response(
                200,
                json=[
                    {
                        "id": "11111111",
                        "roleId": {"scope": "default", "code": "role1"},
                        "name": "role1",
                        "description": "role one",
                    },
                    {
                        "id": "22222222",
                        "roleId": {"scope": "default", "code": "role2"},
                        "name": "role2",
                        "description": "role two",
                    },
                ],
            )
        )
        client = self.client
        # when we attach an exception is thrown
        sut = identity.IdentityRoleRef(id="xyz", name="none_of_those")
        with pytest.raises(RuntimeError):
            sut.attach(client)

    def test_attach_when_http_error(self, respx_mock):
        # given a server which returns a 500
        respx_mock.get("/identity/api/roles").mock(return_value=httpx.Response(500, json={}))
        client = self.client
        # when we attach a http exception is thrown
        sut = identity.IdentityRoleRef(id="xyz", name="none_of_those")
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeIdentityRoleResource:
    client = httpx.Client(base_url=TEST_BASE)

    def test_create(self, respx_mock):
        respx_mock.post("/identity/api/roles").mock(
            return_value=httpx.Response(
                200, json={"id": "aaaaaa", "roleId": {"scope": "default", "code": "role1"}}
            )
        )
        # given a desired role
        sut = identity.IdentityRoleResource(id="role_id", name="role1", description="role one")
        # when we create it
        state = sut.create(self.client)
        # then the state is returned
        assert state == {
            "id": "role_id",
            "scope": "default",
            "name": "role1",
            "code": "role1",
            "roleId": "aaaaaa",
        }
        # and a create request was sent
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/identity/api/roles"
        assert json.loads(request.content) == {"name": "role1", "description": "role one"}

    def test_update_with_no_changes(self, respx_mock):
        respx_mock.get("/identity/api/roles/bxbxbx").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": "bxbxbx",
                    "roleId": {"scope": "default", "code": "role_b"},
                    "source": "source1",
                    "samlName": "abcxxxx",
                    "name": "role_b",
                    "description": "role bee",
                },
            )
        )
        # given a desired role
        sut = identity.IdentityRoleResource(id="res_id", name="role_b", description="role bee")
        old_state = SimpleNamespace(
            roleId="bxbxbx", id="res_id", scope="default", code="role_b", name="role_b"
        )
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and a read was made
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == "/identity/api/roles/bxbxbx"
        # but no PUT

    def test_update_with_change_description(self, respx_mock):
        respx_mock.get("/identity/api/roles/bxbxbx").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": "bxbxbx",
                    "roleId": {"scope": "default", "code": "role_b"},
                    "source": "source1",
                    "samlName": "abcxxxx",
                    "name": "role_b",
                    "description": "role bee",
                },
            )
        )
        respx_mock.put("/identity/api/roles/bxbxbx").mock(return_value=httpx.Response(200, json={}))
        # given a desired role
        sut = identity.IdentityRoleResource(
            id="res_id", name="role_b", description="modified description"
        )
        old_state = SimpleNamespace(
            roleId="bxbxbx", id="res_id", scope="default", code="role_b", name="role_b"
        )
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and the put is sent
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/identity/api/roles/bxbxbx"
        assert json.loads(request.content) == {"description": "modified description"}

    def test_update_with_change_name_should_throw(self, respx_mock):
        # given a desired role
        sut = identity.IdentityRoleResource(id="res_id", name="modified_name", description="role bee")
        old_state = SimpleNamespace(
            roleId="bxbxbx", id="res_id", scope="default", code="role_b", name="role_b"
        )
        # when we update it throws
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_delete(self, respx_mock):
        respx_mock.delete("/identity/api/roles/bxbxbx").mock(return_value=httpx.Response(200, json={}))
        # given a role that exists
        old_state = SimpleNamespace(
            roleId="bxbxbx", id="res_id", scope="default", code="role_b", name="role_b"
        )
        # when we delete it
        identity.IdentityRoleResource.delete(self.client, old_state)
        # then a delete request is sent with the roleId
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/identity/api/roles/bxbxbx"

    def test_deps(self):
        # given a desired role
        sut = identity.IdentityRoleResource(id="res_id", name="modified_name", description="role bee")
        # it's deps are empty
        assert sut.deps() == []


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeUserRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_attach_exists(self, respx_mock):
        # given the user "match" exists in the remote
        respx_mock.get("/identity/api/users").mock(
            return_value=httpx.Response(
                200, json=[{"id": "cvex", "login": "no_match"}, {"id": "bdxe", "login": "match"}]
            )
        )
        # when we attach a UserRef
        sut = identity.UserRef(id="user", login="match")
        sut.attach(self.client)
        # then it's userId will be from the matching entry
        assert sut.user_id == "bdxe"

    def test_attach_when_not_exists(self, respx_mock):
        # given the user "match" does not exist
        respx_mock.get("/identity/api/users").mock(
            return_value=httpx.Response(
                200,
                json=[{"id": "cvex", "login": "no_match"}, {"id": "bdxe", "login": "other_no_match"}],
            )
        )
        # when we attach a UserRef
        sut = identity.UserRef(id="user", login="match")
        # an exception is thrown
        with pytest.raises(RuntimeError):
            sut.attach(self.client)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeUserResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_create(self, respx_mock):
        respx_mock.post("/identity/api/users").mock(
            return_value=httpx.Response(200, json={"id": "bcdef"})
        )
        # given a desired user
        sut = identity.UserResource(
            id="user",
            login="match",
            first_name="Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        # when we create
        sut.create(self.client)
        # then the request is made
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/identity/api/users"
        assert json.loads(request.content) == {
            "login": "match",
            "firstName": "Jess",
            "lastName": "Blofeldt",
            "emailAddress": "jess@blo.com",
            "type": "Service",
        }
        # and the user gets an id
        assert sut.user_id == "bcdef"

    def test_deps(self):
        # given a desired user
        sut = identity.UserResource(
            id="user",
            login="match",
            first_name="Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        # it has no dependencies
        assert sut.deps() == []

    def test_update_with_change(self, respx_mock):
        # given a user in the remote
        respx_mock.get("/identity/api/users/xxyyzz").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": "xxyyzz",
                    "login": "login@login.com",
                    "firstName": "Jess",
                    "lastName": "Blofeldt",
                    "emailAddress": "jess@blo.com",
                    "type": "Service",
                    "status": "active",
                },
            )
        )
        respx_mock.put("/identity/api/users/xxyyzz").mock(return_value=httpx.Response(200, json={}))
        # and existing state
        old_state = SimpleNamespace(id="22", userId="xxyyzz")
        # and desired state with a different firstname
        sut = identity.UserResource(
            id="22",
            login="login@login.com",
            first_name="New Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        # when we update
        new_state = sut.update(self.client, old_state)
        # then the request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/identity/api/users/xxyyzz"
        assert json.loads(request.content) == {
            "login": "login@login.com",
            "firstName": "New Jess",
            "lastName": "Blofeldt",
            "emailAddress": "jess@blo.com",
            "type": "Service",
        }
        # and the new state is returned (the same as before)
        assert new_state is not None
        assert new_state["userId"] == "xxyyzz"

    def test_update_with_no_change(self, respx_mock):
        # given a user in the remote
        respx_mock.get("/identity/api/users/xxyyzz").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": "xxyyzz",
                    "login": "login@login.com",
                    "firstName": "Jess",
                    "lastName": "Blofeldt",
                    "emailAddress": "jess@blo.com",
                    "type": "Service",
                    "status": "active",
                    "external": False,
                },
            )
        )
        # and existing state
        old_state = SimpleNamespace(id="22", userId="xxyyzz")
        # and desired state which is the same as the remote
        sut = identity.UserResource(
            id="22",
            login="login@login.com",
            first_name="Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        # when we update
        new_state = sut.update(self.client, old_state)
        # the new state is None
        assert new_state is None
        # and no put request is made

    def test_update_cannot_change_login(self, respx_mock):
        # given a user in the remote
        respx_mock.get("/identity/api/users/xxyyzz").mock(
            return_value=httpx.Response(
                200,
                json={
                    "id": "xxyyzz",
                    "login": "login@login.com",
                    "firstName": "Jess",
                    "lastName": "Blofeldt",
                    "emailAddress": "jess@blo.com",
                    "type": "Service",
                    "status": "active",
                    "external": False,
                },
            )
        )
        # and existing state
        old_state = SimpleNamespace(id="22", userId="xxyyzz")
        # and desired state with a different login
        sut = identity.UserResource(
            id="22",
            login="different_login@login.com",
            first_name="Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        with pytest.raises(RuntimeError):
            sut.update(self.client, old_state)

    def test_delete(self, respx_mock):
        respx_mock.delete("/identity/api/users/xxyyzz").mock(return_value=httpx.Response(200, json={}))
        # given a role that exists
        old_state = SimpleNamespace(userId="xxyyzz", id="res_id")
        # when we delete it
        identity.UserResource.delete(self.client, old_state)
        # then a delete request is sent with the userId
        request = respx_mock.calls.last.request
        assert request.method == "DELETE"
        assert request.url.path == "/identity/api/users/xxyyzz"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeRoleAssignment:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @pytest.fixture
    def created_role(self, respx_mock):
        respx_mock.post("/identity/api/roles").mock(
            return_value=httpx.Response(
                200, json={"id": "role01", "roleId": {"scope": "default", "code": "role1"}}
            )
        )
        # given a desired role
        role = identity.IdentityRoleResource(id="role_id", name="role1", description="role one")
        role.create(self.client)
        return role

    @pytest.fixture
    def created_user(self, respx_mock):
        respx_mock.post("/identity/api/users").mock(
            return_value=httpx.Response(200, json={"id": "user02"})
        )
        user = identity.UserResource(
            id="user",
            login="match",
            first_name="Jess",
            last_name="Blofeldt",
            email_address="jess@blo.com",
            type=identity.UserType.SERVICE,
        )
        user.create(self.client)
        return user

    def test_create(self, respx_mock, created_user, created_role):
        respx_mock.put("/identity/api/roles/role01/users/user02").mock(
            return_value=httpx.Response(200, json={})
        )
        # given a user and a role
        # and the desired assignment
        sut = identity.RoleAssignment(id="ass1", user=created_user, role=created_role)
        # when we create
        sut.create(self.client)
        # then the request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/identity/api/roles/role01/users/user02"

    def test_create_with_user_ref(self, respx_mock, created_role):
        respx_mock.get("/identity/api/users").mock(
            return_value=httpx.Response(200, json=[{"id": "user02", "login": "match"}])
        )
        user_ref = identity.UserRef(id="user", login="match")
        user_ref.attach(self.client)

        respx_mock.put("/identity/api/roles/role01/users/user02").mock(
            return_value=httpx.Response(200, json={})
        )

        # given a user ref and a role
        # and the desired assignment
        sut = identity.RoleAssignment(id="ass1", user=user_ref, role=created_role)
        # when we create
        sut.create(self.client)
        # then the request is made
        request = respx_mock.calls.last.request
        assert request.method == "PUT"
        assert request.url.path == "/identity/api/roles/role01/users/user02"

    def test_update_no_change(self, respx_mock, created_user, created_role):
        # gvien an assignment which matches the remote
        sut = identity.RoleAssignment(id="ass1", user=created_user, role=created_role)
        # when we update
        old_state = SimpleNamespace(id="ass1", roleId="role01", userId="user02")
        new_state = sut.update(self.client, old_state)
        # the new state is none
        assert new_state is None
        # and no put request is made

    def test_update_change_user_when_remote_exists(self, respx_mock, created_user, created_role):
        respx_mock.put("/identity/api/roles/role01/users/user02").mock(
            return_value=httpx.Response(200, json={})
        )
        respx_mock.delete("/identity/api/roles/role01/users/user01").mock(
            return_value=httpx.Response(200, json={})
        )
        # given the existing remote is for user01
        old_state = SimpleNamespace(id="ass1", roleId="role01", userId="user01")
        # and a desired state of user02
        sut = identity.RoleAssignment(id="ass1", user=created_user, role=created_role)
        # when we update
        new_state = sut.update(self.client, old_state)
        # then the new state is user02
        assert new_state == {"id": "ass1", "roleId": "role01", "userId": "user02"}
        # and the existing assignment is deleted
        request = respx_mock.calls[-2].request
        assert request.method == "DELETE"
        assert request.url.path == "/identity/api/roles/role01/users/user01"
        # and a new one is created
        request = respx_mock.calls[-1].request
        assert request.method == "PUT"
        assert request.url.path == "/identity/api/roles/role01/users/user02"

    def test_deps(self, created_user, created_role):
        # given a desired assignment
        sut = identity.RoleAssignment(id="ass1", user=created_user, role=created_role)
        # it depends on the user and the role
        assert sut.deps() == [created_user, created_role]
        ref = identity.UserRef(id="user", login="match")
        sut = identity.RoleAssignment(id="ass1", user=ref, role=created_role)
        assert sut.deps() == [ref, created_role]

    def test_delete(self, respx_mock, created_user, created_role):
        respx_mock.delete("/identity/api/roles/role01/users/user02").mock(
            return_value=httpx.Response(200, json={})
        )
        # given an existing remote
        old_state = SimpleNamespace(id="ass1", roleId="role01", userId="user02")
        # when we delete it
        identity.RoleAssignment.delete(self.client, old_state)
        # the existing assignment is deleted
        request = respx_mock.calls[-1].request
        assert request.method == "DELETE"
        assert request.url.path == "/identity/api/roles/role01/users/user02"



from __future__ import annotations

import hashlib
import json
from pathlib import PurePath
from types import SimpleNamespace
from typing import Annotated, Any

import httpx
from pydantic import AfterValidator, BaseModel, ConfigDict, Field, computed_field, model_serializer

from .resource_abc import Resource, register_resource


def validate_recipe(v: dict[str, Any]) -> dict[str, Any]:
    if v is None:
        return v
    if isinstance(v, PurePath) and not v.is_absolute():
        raise ValueError("Path should be an absolute path")
    return v


@register_resource()
class RecipeResource(BaseModel, Resource):
    """
    Manage a Recipe resource.

    Attributes:

        id: str
            The unique identifier of this resource.
        scope: str
            The scope of the recipe.
        code: str
            The code of the recipe.
        recipe: dict[str, Any] | PurePath | None
            The recipe definition.
            If a dictionary, it is a recipe as defined in the LUSID API.
            If a PurePath, it is a path to a JSON file containing the recipe.

    Example:

        .. code-block:: python

            from fbnconfig import recipe

            recipe = recipe.RecipeResource(
                scope="my_scope",
                code="my_code",
                recipe={
                    "market": {...},
                }
            )
    """

    model_config = ConfigDict(arbitrary_types_allowed=True)
    id: str = Field(init=True, exclude=True)
    scope: str = Field(init=True, exclude=True)
    code: str = Field(init=True, exclude=True)
    recipe: Annotated[dict[str, Any] | PurePath, AfterValidator(validate_recipe)]

    @computed_field()
    def _recipe_model_(self) -> dict[str, Any]:
        recipe_model = {}
        if isinstance(self.recipe, PurePath):
            with open(self.recipe, "rb") as ff:
                recipe_model = json.loads(ff.read())
        else:
            recipe_model = self.recipe
        recipe_model.pop("scope", None)
        recipe_model.pop("code", None)
        return recipe_model

    @model_serializer(when_used="always")
    def serialize_model(self) -> dict[str, Any]:
        return {
            "configurationRecipe":
            # Operator "|" not supported for types "dict[str, str]" and "() -> dict[str, Any]
            {"scope": self.scope, "code": self.code} | self._recipe_model_  # type:ignore
        }

    def read(self, client, old_state):
        return self.__get_recipe__(client, old_state.scope, old_state.code, as_at=None)

    def create(self, client: httpx.Client):
        source = self.model_dump(mode="json", exclude_none=True)
        as_at = client.post("/api/api/recipes", json=source).json()["value"]
        remote = self.__get_recipe__(client, self.scope, self.code, as_at)

        return {
            "scope": self.scope,
            "code": self.code,
            "source_version": self.__get_hash__(source),
            "remote_version": self.__get_hash__(remote),
        }

    def update(self, client, old_state):
        if [self.scope, self.code] != [old_state.scope, old_state.code]:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        dump = self.model_dump(mode="json", exclude_none=True)
        remote_version = self.__get_hash__(remote)
        source_version = self.__get_hash__(dump)
        if source_version == old_state.source_version and remote_version == old_state.remote_version:
            return None

        return self.create(client)

    @staticmethod
    def delete(client, old_state: SimpleNamespace):
        client.delete(f"/api/api/recipes/{old_state.scope}/{old_state.code}")

    def deps(self) -> list[Any]:
        return []

    @staticmethod
    def __get_recipe__(client: httpx.Client, scope: str, code: str, as_at: str | None) -> dict[str, Any]:
        r = client.get(
            f"/api/api/recipes/{scope}/{code}", params={"asAt": as_at} if as_at else None
        ).json()["value"]
        r.pop("scope", None)
        r.pop("code", None)
        return r

    @staticmethod
    def __get_hash__(content) -> str:
        return hashlib.sha256(json.dumps(content, sort_keys=True).encode()).hexdigest()



from fbnconfig import Deployment, scheduler

"""
An example configuration for defining scheduler related entities.
The script configures the following entities:
- Image
- Job
- Schedule

More information can be found here:
https://support.lusid.com/knowledgebase/article/KA-01698/
https://support.lusid.com/knowledgebase/article/KA-02048/
https://support.lusid.com/knowledgebase/article/KA-02049/
"""


def configure(env):
    tag = "s17"

    image = scheduler.ImageResource(
        id="img1", source_image="docker.io/alpine:latest", dest_name="schedule-example", dest_tag=tag
    )

    job = scheduler.JobResource(
        id="job1",
        scope="sc1",
        code="job",
        image=image,
        name="Example Job",
        description="Example job resource",
        min_cpu="1",
        max_cpu="2",
        argument_definitions={
            "arg1": scheduler.EnvironmentArg(
                data_type="String",
                required=False,
                description="Example argument",
                order=1,
                default_value="default-value",
            )
        },
    )

    schedule = scheduler.ScheduleResource(
        id="sch1",
        name="schedule",
        scope="sc1",
        code="python-schedule",
        expression="0 30 4 * * ? *",
        timezone="Europe/London",
        job=job,
        description="Example schedule",
        arguments={"arg1": "none-default-value"},
        enabled=False,
    )

    return Deployment("scheduler_example", [image, job, schedule])



import os
from types import SimpleNamespace

from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
import tests.integration.access as access
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("access")
    print(f"\nRunning for deployment {deployment_name}...")

    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    deployment_name = setup_deployment.name
    client = fbnconfig.create_client(lusid_env, token)
    # setup deployment
    fbnconfig.deploy(access.configure(setup_deployment), lusid_env, token)
    # check it exists
    client.request("get", f"/access/api/policies/policy-{deployment_name}")
    # Tear it down
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)
    # check it was deleted
    try:
        client.request("get", f"/access/api/policies/policy-{deployment_name}")
    except HTTPStatusError as error:
        assert error.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deploy(access.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    search = client.request("get", f"/access/api/roles/role-{setup_deployment.name}")
    assert search.status_code == 200


def test_update_nochange(setup_deployment):
    fbnconfig.deploy(access.configure(setup_deployment), lusid_env, token)
    update = fbnconfig.deploy(access.configure(setup_deployment), lusid_env, token)
    assert [a.change for a in update] == ["nochange"] * 2 + ["attach"] + ["nochange"] * 5



import json
import textwrap
from types import SimpleNamespace
from typing import Any, Dict, List, Optional
from unittest.mock import patch

import httpx
import pytest
from respx import Route, Router, mock

from fbnconfig import lumi
from fbnconfig.backoff_handler import BackoffHandler
from fbnconfig.lumi import TaskStatus

TEST_BASE = "https://foo.lusid.com"


def response_hook(response: httpx.Response):
    response.raise_for_status()


@pytest.fixture(scope="module", autouse=True)
def dont_sleep():
    with patch("time.sleep", return_value=None) as _fixture:
        yield _fixture


def setup_sql_background_routes(
    respx_mock: Router,
    content: Optional[str] = None,
    execution_id: str = "abcdefg",
    task_status: TaskStatus = TaskStatus.RAN_TO_COMPLETION,
    status_code: Optional[List[int]] = None,
    result_data: Optional[Any] = None,
) -> dict[str, Route]:
    routes = {
        "put": respx_mock.put(
            "/honeycomb/api/SqlBackground",
            # # don't pass values as arguments if they're set to none
            **dict(
                filter(
                    lambda item: item[1] is not None,
                    {"content": content, "headers": [("Content-type", "text/plain")]}.items(),
                )
            ),
        ).mock(
            side_effect=[
                httpx.Response(
                    status_code.pop() if status_code else 202, json={"executionId": execution_id}
                )
            ]
        ),
        "status": respx_mock.get(f"/honeycomb/api/SqlBackground/{execution_id}").mock(
            side_effect=[
                httpx.Response(status_code.pop() if status_code else 200, json={"status": task_status})
            ]
        ),
        "result": respx_mock.get(f"/honeycomb/api/SqlBackground/{execution_id}/jsonproper").mock(
            side_effect=[httpx.Response(status_code.pop() if status_code else 200, json=result_data)]
        ),
    }

    return routes


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeViewRef:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})
    view_attach_content = textwrap.dedent("""\
                select r.Version from sys.registration as r
                where name = 'views.test.view1'
                order by r.Version asc
                limit 1
            """)

    def test_attach_when_exists(self, respx_mock, monkeypatch):
        # given a view in lumi
        setup_sql_background_routes(
            respx_mock, content=self.view_attach_content, result_data=["one row"]
        )

        client = self.client
        # when we attach
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        sut.attach(client)
        assert sut._version == "one row"

    def test_attach_when_not_exists(self, respx_mock):
        setup_sql_background_routes(respx_mock, content=self.view_attach_content, result_data=[])
        # given no matching views exist yet

        client = self.client
        # when we attach a view ref
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        # then it raises
        with pytest.raises(RuntimeError) as ex:
            sut.attach(client)
        assert "Failed to attach ref to views.test.view1" in str(ex.value)

    @mock(assert_all_called=False)
    def test_attach_when_http_error(self, respx_mock):
        # given the query fails
        routes: dict[str, Route] = setup_sql_background_routes(
            respx_mock, content=self.view_attach_content, status_code=[400]
        )

        client = self.client
        # when we attach a view ref
        sut = lumi.ViewRef(id="abc", provider="views.test.view1")
        # http error is raised
        with pytest.raises(httpx.HTTPStatusError):
            sut.attach(client)
        assert routes["put"].call_count == 1


# constructs the row from sys.file
def make_sys_file(sql: str, metadata: Dict, version: int) -> Dict:
    return {"Content": f"{sql}--- MetaData --- " + json.dumps(metadata), "Version": version}


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeViewResource:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def setup_class():
        # reduce time waiting for registration
        lumi.ViewResource.Registration.tries = 3
        lumi.ViewResource.Registration.wait_time = 0

    def test_create(self, respx_mock):
        exec_id = "abcdefg"

        expected = textwrap.dedent("""\
                   @x = use Sys.Admin.SetupView
                   --provider="Views.Unit.Test"
                   --description="a view"
                   ----
                   select * from foo;
                   enduse;
                   select * from @x;
               """)
        # setup_sql_background_routes(respx_mock, content=expected, result_data={})

        expected_registration = textwrap.dedent("""\
                    select Version from sys.registration where Name='Views.Unit.Test'
                    order by Version asc
                    limit 1
                """)

        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )
        # setup_sql_background_routes(
        #     respx_mock,
        #     content=expected_registration,
        #     execution_id="123",
        #     result_data=[{"Version": 1}])

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        # the state is returned
        respx_mock.assert_all_called()
        assert put.calls[0].request.content.decode("utf8") == expected
        assert put.calls[1].request.content.decode("utf8") == expected_registration
        assert state == {"provider": "Views.Unit.Test"}

    def test_create_waits_for_registration(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                # Query creation + 3x registration
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})
            ]
            * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={})]
            +
            # and the registration check was performed a second time
            # because no version came back on the first check
            [httpx.Response(200, json=[])] * 2
            + [httpx.Response(200, json=[{"Version": "1"}])]
        )

        expected = textwrap.dedent("""\
                    select Version from sys.registration where Name='Views.Unit.Test'
                    order by Version asc
                    limit 1
                """)

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        respx_mock.assert_all_called()
        assert put.calls[1].request.content.decode("utf8") == expected
        # the state is returned
        assert state == {"provider": "Views.Unit.Test"}

    def test_create_completes_if_registration_incomplete(self, respx_mock, capsys):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                # Query creation + 3x registration
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})
            ]
            * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={})] + [httpx.Response(200, json=[])] * 3
        )

        # given a view to create
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo;", description="a view"
        )
        # when we do
        state = sut.create(self.client)
        # the state is returned even though the registration did not
        # complete in time
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was performed 3 times
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        respx_mock.assert_all_called()
        assert put.calls[1].request.content.decode("utf8") == expected

        # check warning is returned
        captured = capsys.readouterr()
        assert captured.err == (
            f"warning: no view registration after "
            f"{lumi.ViewResource.Registration.tries}"
            f" tries for Views.Unit.Test"
        )

    def test_create_with_saved_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired view with one of the saved options
        # (ones which impact the stored view in lumi)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            documentation_link="http://example.com",
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # and a create request was sent with an =
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="http://example.com"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_with_saved_bool_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired state with a boolean saved option
        # variableShape
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            variable_shape=True,
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the create adds the option without an =
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --variableShape
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_with_test_option(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given desired includes limit, an option which is used for
        # running the view but does not get saved
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            limit=42,
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the test option gets sent and is not quoted because
        # it's an int
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --limit=42
            --description="a view"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=23,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter_escaped(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a text parameter where the value
        # includes a double quote
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Text,
                    value='embedded " quote',
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the text is wrapped in quotes and the embedded quote
        # gets escaped
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Text,"embedded "" quote",false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_parameter_no_tooltip(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1", type=lumi.ParameterType.Int, value=23, set_as_default_value=False
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_scalar_variable(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a variable which is the default value
        # for a parameter
        p1_value = lumi.Variable(name="p1_value", type=lumi.VariableType.Scalar, sql="select 1 + 2")
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo;",
            description="a view",
            variables=[p1_value],
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=p1_value,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @@p1_value = select 1 + 2;
            @x = use Sys.Admin.SetupView with @@p1_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,@@p1_value,false,"p1 is nice"
            ----
            select * from foo;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_table_variable(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a table variable used as a
        # parameter
        p2_value = lumi.Variable(
            name="p2_value",
            type=lumi.VariableType.Table,
            sql="""select * from ( values(1, 2, 3), (4, 5, 6))""",
        )
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from @p2;",
            description="a view",
            variables=[p2_value],
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Table,
                    value=p2_value,
                    is_mandatory=False,
                    tooltip="p2 is nice",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the parameter is added at the end of the options
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @p2_value = select * from ( values(1, 2, 3), (4, 5, 6));
            @x = use Sys.Admin.SetupView with @p2_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Table,@p2_value,false,"p2 is nice"
            ----
            select * from @p2;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_create_two_variables(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[httpx.Response(200, json={}), httpx.Response(200, json=[{"Version": 1}])]
        )

        # given a desired with a one table and one scalar
        # for a parameter
        p1_value = lumi.Variable(name="p1_value", type=lumi.VariableType.Scalar, sql="select 1 + 2")
        p2_value = lumi.Variable(
            name="p2_value",
            type=lumi.VariableType.Table,
            sql="""select * from ( values(1, 2, 3), (4, 5, 6))""",
        )
        # the table is used as a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from @p2;",
            description="a view",
            variables=[p1_value, p2_value],
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Table,
                    value=p2_value,
                    is_mandatory=False,
                    tooltip="p2 is nice",
                )
            ],
        )
        # when we create it
        state = sut.create(self.client)
        assert state == {"provider": "Views.Unit.Test"}
        # the vars are initialised, the with statement includes them
        # and the table parameter is included
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @@p1_value = select 1 + 2;
            @p2_value = select * from ( values(1, 2, 3), (4, 5, 6));
            @x = use Sys.Admin.SetupView with @@p1_value, @p2_value
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Table,@p2_value,false,"p2 is nice"
            ----
            select * from @p2;
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_no_changes(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})])
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                )
            ]
        )

        # and a desired state that is the same;
        # new line at the end of script should not count as change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="""select * from foo
            """,
            description="a view",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and the read query was sent
        request = put.calls[0].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            select f.Content, r.Version from sys.file as f
            join sys.registration as r on r.Name = 'Views.Unit.Test'
            where path = 'databaseproviders/Views/Unit/Test.sql'
            order by r.Version asc
            limit 1
        """)
        assert request.content.decode() == expected
        # and we read the remote but didn't write anything
        assert len(respx_mock.calls) == 3

    def test_update_with_test_option_means_no_change(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})]
        )
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                )
            ]
        )

        # and a desired that is the same except we add test_options
        # limit, useDryRun, groupby, filter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            limit=42,
            use_dry_run=True,
            group_by="one",
            filter="something",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None because test_options are not saved
        assert state is None

    def test_update_with_extra_saved_option(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, registration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and a desired adds a saved option (documentationLink)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="http://example.com"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_does_registration_check(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 2xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 2}]),  # registration complete
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was performed twice
        # because the version before update was 1 and we wait until it's increased
        check2 = put.calls[3].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check2.content.decode("utf8") == expected

    def test_update_completes_when_registration_version_does_not_update(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 3xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # register within the 3 tries
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a state is returned even though the reg check didn't complete
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was 3 times
        # because the version didn't update
        check = put.calls[4].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_update_completes_when_no_registration_rows_returned(self, respx_mock):
        # given an existing remote where the view is not going to
        # register within the 3 tries
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(
            # Read, update, 3xregistration check
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 5
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[{"Version": 1}]),  # registration false
                httpx.Response(200, json=[]),  # no rows returned
            ]
        )

        # and a desired makes a change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="http://example.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a state is returned even though the reg check didn't complete
        assert state == {"provider": "Views.Unit.Test"}
        # and the registration check was 3 times
        # because the version didn't update
        check = put.calls[4].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_update_defaulted_saved_option(self, respx_mock):
        # given an existing View created without the variableShape option
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [],
                                "IsWithinDirectProviderView": False,
                            },
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and the desired leaves it empty
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then no change is detected because False is a server default
        # and the user has not specified it
        assert state is None

    def test_update_with_modified_saved_option(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {"Description": "a view", "Parameters": [], "DocumentationLink": "bar.com"},
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )

        # and a desired adds a saved option (documentationLink)
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            documentation_link="foo.com",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --documentationLink="foo.com"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_modified_sql(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired changes the sql
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from something_else",
            description="a view",
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then a change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            ----
            select * from something_else
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_extra_parameter(self, respx_mock):
        # given an existing remote
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo", {"Description": "a view", "Parameters": []}, version=1
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired that is the same but adds a parameter
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=23,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the new parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we update the view with the parameter
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,23,false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_removed_parameter(self, respx_mock):
        # given an existing view with one parameter
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [{"Name": "existing", "Type": "Text"}],
                            },
                            version=1,
                        )
                    ],
                ),
                # update
                httpx.Response(200, json=[]),
                # registration check
                httpx.Response(200, json=[{"Version": 2}]),
            ]
        )
        # and a desired that removes the parameter
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we created a new view with none
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_modified_parameter(self, respx_mock):
        # given an existing view with integer parameter p1
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {"Description": "a view", "Parameters": [{"Name": "p1", "Type": "Int"}]},
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired that changes the parameter type to text
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Text,
                    value="some text",
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view with a text parameter
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Text,"some text",false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_useasdefault_changes_value(self, respx_mock):
        # given an existing view with integer parameter p1
        # with a default of zero
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Name": "p1",
                                        "Type": "Int",
                                        "DefaultValue": 0,
                                        "Description": "p1 is nice",
                                    }
                                ],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired with setAsDefault=true which means to use the value as a default
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=20,
                    set_as_default_value=True,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view with the new value
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p1,Int,20,true,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_useasdefault_same_value(self, respx_mock):
        # given an existing view with integer parameter p1
        # with a default of 20
        exec_id = "abcdefg"
        respx_mock.put("/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]).mock(
            side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Name": "p1",
                                        "Type": "Int",
                                        "DefaultValue": 20,
                                        "Description": "p1 is nice",
                                    }
                                ],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired with setAsDefault but a value also 20
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    type=lumi.ParameterType.Int,
                    value=20,
                    set_as_default_value=True,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then there is no change
        assert state is None

    def test_update_with_renamed_parameter(self, respx_mock):
        # given an existing view with integer parameter p1
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # read
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from foo",
                            {
                                "Description": "a view",
                                "Parameters": [{"Name": "p1", "Type": "Int", "ExtraField": "boo"}],
                            },
                            version=1,
                        )
                    ],
                ),
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 2}]),  # registration check
            ]
        )
        # and a desired renames p1 to p2
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p2",
                    type=lumi.ParameterType.Int,
                    value=12,
                    set_as_default_value=False,
                    tooltip="p1 is nice",
                )
            ],
        )
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the parameter change is detected
        assert state == {"provider": "Views.Unit.Test"}
        # and we modify the view to only have p2
        request = put.calls[1].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.Test"
            --description="a view"
            --parameters
            p2,Int,12,false,"p1 is nice"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_provider_name(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[]),  # de-reg complete
                httpx.Response(200, json=[]),  # update
                httpx.Response(200, json=[{"Version": 1}]),  # reg complete
            ]
        )

        # given a view which has been created with a provider name
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we change the provider name
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.NotTheSame", sql="select * from foo", description="a view"
        )
        # and update
        sut.update(self.client, old_state)
        # then the existing provider gets deleted
        first = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider=Views.Unit.Test
            --deleteProvider
            ----
            select 1 as deleting
            enduse;
            select * from @x;
        """)
        assert first.content.decode() == expected
        # and a new one gets created
        request = put.calls[2].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider="Views.Unit.NotTheSame"
            --description="a view"
            ----
            select * from foo
            enduse;
            select * from @x;
        """)
        assert request.content.decode("utf8") == expected

    def test_update_with_no_changes_and_table_param(self, respx_mock):
        # given an existing remote

        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})])

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})]
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(
                    200,
                    json=[
                        make_sys_file(
                            "select * from #PARAMETERVALUE(p1);",
                            {
                                "Description": "a view",
                                "Parameters": [
                                    {
                                        "Columns": [
                                            {
                                                "IsPrimaryKey": "false",
                                                "IsMain": "false",
                                                "IsRequiredByProvider": "false",
                                                "MandatoryForActions": None,
                                                "ClientIds": None,
                                                "Name": "3 + 4",
                                                "Type": "BigInt",
                                                "Description": None,
                                                "DisplayName": "3 + 4",
                                                "ConditionUsage": 0,
                                                "SampleValues": None,
                                                "AllowedValues": None,
                                            }
                                        ],
                                        "Name": "p1",
                                        "Type": "Table",
                                        "Description": "I am a description"
                                        "\nAvailable columns:\n3 + 4 (BigInt)",
                                        "DisplayName": "p1",
                                        "ConditionUsage": 0,
                                    },
                                    {
                                        "Columns": [
                                            {
                                                "IsPrimaryKey": "false",
                                                "IsMain": "false",
                                                "IsRequiredByProvider": "false",
                                                "MandatoryForActions": None,
                                                "ClientIds": None,
                                                "Name": "5 + 6",
                                                "Type": "BigInt",
                                                "Description": None,
                                                "DisplayName": "5 + 6",
                                                "ConditionUsage": 0,
                                                "SampleValues": None,
                                                "AllowedValues": None,
                                            }
                                        ],
                                        "Name": "p2",
                                        "Type": "Table",
                                        "Description": "\nAvailable columns:\n5 + 6 (BigInt)",
                                        "DisplayName": "p2",
                                        "ConditionUsage": 0,
                                    },
                                ],
                            },
                            version=1,
                        )
                    ],
                )
            ]
        )

        p1_value = lumi.Variable(name="p1", type=lumi.VariableType.Table, sql="select * from 3 + 4")

        p2_value = lumi.Variable(name="p2", type=lumi.VariableType.Table, sql="select a,b,c from Test")
        # and a desired state that is the same;
        # new line at the end of script should not count as change
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="""select * from #PARAMETERVALUE(p1);""",
            description="a view",
            parameters=[
                lumi.Parameter(
                    name="p1",
                    value=p1_value,
                    is_mandatory=False,
                    tooltip="I am a description",
                    type=lumi.ParameterType.Table,
                ),
                lumi.Parameter(
                    name="p2",
                    value=p2_value,
                    is_mandatory=False,
                    type=lumi.ParameterType.Table,
                    tooltip="",
                ),
            ],
        )

        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we update it
        state = sut.update(self.client, old_state)
        # then the state is None
        assert state is None
        # and the read query was sent
        request = put.calls[0].request
        assert request.method == "PUT"
        expected = textwrap.dedent("""\
               select f.Content, r.Version from sys.file as f
               join sys.registration as r on r.Name = 'Views.Unit.Test'
               where path = 'databaseproviders/Views/Unit/Test.sql'
               order by r.Version asc
               limit 1
           """)
        assert request.content.decode() == expected
        # and we read the remote but didn't write anything
        assert len(respx_mock.calls) == 3

    @staticmethod
    def test_deps_none():
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="select * from foo", description="a view"
        )
        assert sut.deps() == []

    @staticmethod
    def test_deps_extra():
        sut = lumi.ViewResource(
            id="my-view",
            provider="Views.Unit.Test",
            sql="select * from foo",
            description="a view",
            dependencies=[1, 2, 3],
        )
        assert sut.deps() == [1, 2, 3]

    def test_delete(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 2)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 2
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                # delete
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[]),  # de-registration check
            ]
        )
        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )

        # given a view that exists in the remote
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # then sql to remove the view
        request = put.calls[0].request
        expected = textwrap.dedent("""\
            @x = use Sys.Admin.SetupView
            --provider=Views.Unit.Test
            --deleteProvider
            ----
            select 1 as deleting
            enduse;
            select * from @x;
        """)
        assert request.content.decode() == expected

    def test_delete_does_registration_check(self, respx_mock):
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 3)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 3
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[]),  # deregistered
            ]
        )

        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )

        # given a view that exists in the remote
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # the registration check was performed twice because we waited for
        # the view to be removed
        check = put.calls[2].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected

    def test_delete_completes_when_reg_incomplete(self, respx_mock, capsys):
        # given de-registration takes more than 3 tries
        exec_id = "abcdefg"
        put: Route = respx_mock.put(
            "/honeycomb/api/SqlBackground", headers=[("Content-type", "text/plain")]
        ).mock(side_effect=[httpx.Response(202, json={"executionId": exec_id})] * 4)

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION})] * 4
        )

        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}/jsonproper").mock(
            side_effect=[
                httpx.Response(200, json=[]),  # delete
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
                httpx.Response(200, json=[{"Version": 5}]),  # still registered
            ]
        )

        sut = lumi.ViewResource(
            id="my-view", provider="Views.Unit.Test", sql="""select * from 2+2""", description="a view"
        )
        # given a view that exists in the remnte
        old_state = SimpleNamespace(provider="Views.Unit.Test")
        # when we delete it
        sut.delete(self.client, old_state)
        # a delete was sent and the registration check was performed 3 times
        check = put.calls[3].request
        expected = textwrap.dedent("""\
            select Version from sys.registration where Name='Views.Unit.Test'
            order by Version asc
            limit 1
        """)
        assert check.content.decode("utf8") == expected
        # check warning is returned
        captured = capsys.readouterr()
        assert captured.err == (
            f"warning: no view deregistration after "
            f"{lumi.ViewResource.Registration.tries}"
            f" tries for Views.Unit.Test"
        )


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeBackoffHandler:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    @staticmethod
    def test_pause_time_smaller_than_max_pause_time():
        backoff_handler_test = BackoffHandler(pause_time=0.5, max_pause_time=20, beta=1.0001)
        backoff_handler_test.sleep()
        pytest.approx(0.50005, backoff_handler_test.pause_time, 0.00001)

    @staticmethod
    def test_pause_time_equals_max_pause_time():
        backoff_handler_test = BackoffHandler(pause_time=2, max_pause_time=2, beta=1.001)
        backoff_handler_test.sleep()
        assert backoff_handler_test.pause_time == 2

    @staticmethod
    def test_pause_time_bigger_than_max_pause_time():
        with pytest.raises(ValueError) as ex:
            BackoffHandler(pause_time=100, max_pause_time=20, beta=1.1)
            assert "Pause time must be between 0.1 and 20, both inclusive." in str(ex.value)

    @staticmethod
    def test_default_optional_values():
        backoff_handler_test = BackoffHandler()
        backoff_handler_test.sleep()
        pytest.approx(0.1005012520859401, backoff_handler_test.pause_time, 0.00001)

    @staticmethod
    def test_call_1000_times_pause_time_equals_max_pause_time():
        backoff_handler_test = BackoffHandler()
        # I have checked that this is constant after 922 requests
        for i in range(1000):
            backoff_handler_test._update_pause_time()
        assert backoff_handler_test.pause_time == 10


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeQuery:
    client = httpx.Client(base_url=TEST_BASE, event_hooks={"response": [response_hook]})

    def test_wait_for_background_switch_case_correct(self, respx_mock):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": TaskStatus.WAITING_TO_RUN}),
                httpx.Response(200, json={"status": TaskStatus.CREATED}),
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_CHILDREN_TO_COMPLETE}),
                httpx.Response(200, json={"status": TaskStatus.RUNNING}),
                httpx.Response(200, json={"status": TaskStatus.RAN_TO_COMPLETION}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is True

    @pytest.mark.parametrize("status", [TaskStatus.FAULTED, TaskStatus.CANCELED])
    def test_wait_for_background_status_faulted_throws(self, respx_mock, status):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": status}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        with pytest.raises(RuntimeError) as ex:
            lumi.wait_for_background(client=self.client, execution_id=exec_id)
        assert str(ex.value) == f"Query was {status}. Execution id: abcdefg"

    def test_wait_for_background_status_unknown_throws(self, respx_mock):
        exec_id = "abcdefg"
        respx_mock.get(f"/honeycomb/api/SqlBackground/{exec_id}").mock(
            side_effect=[
                httpx.Response(200, json={"status": TaskStatus.WAITING_FOR_ACTIVATION}),
                httpx.Response(200, json={"status": "someunknownstatus"}),
            ]
        )
        assert lumi.wait_for_background(client=self.client, execution_id=exec_id) is False
        with pytest.raises(RuntimeError) as ex:
            lumi.wait_for_background(client=self.client, execution_id=exec_id)
        assert str(ex.value) == "Unknown status: someunknownstatus. Execution id: abcdefg"



from __future__ import annotations

import hashlib
import json
from enum import StrEnum
from types import SimpleNamespace
from typing import Any, Dict, Sequence

import httpx
from pydantic import BaseModel, Field

from .resource_abc import Ref, Resource, register_resource


class Visibility(StrEnum):
    PERSONAL = "personal"
    SHARED = "shared"


class WorkspaceRef(BaseModel, Ref):
    id: str = Field(exclude=True, init=True)
    visibility: Visibility
    name: str

    def attach(self, client):
        try:
            client.request(
                "get", f"/api/api/workspaces/{self.visibility.value}/{self.name}"
            )
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"Workspace {self.visibility.value}/{self.name} not found"
                )
            else:
                raise ex


@register_resource()
class WorkspaceResource(BaseModel, Resource):
    """Manage a workspace"""

    id: str = Field(exclude=True)
    visibility: Visibility
    name: str
    description: str

    def read(
        self, client: httpx.Client, old_state: SimpleNamespace
    ) -> None | Dict[str, Any]:
        workspace_path = f"{old_state.visibility}/{old_state.name}"
        response = client.get(f"/api/api/workspaces/{workspace_path}").json()
        for key in ["links", "version"]:
            response.pop(key, None)
        return response

    def create(self, client: httpx.Client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"visibility"}, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        client.request("POST", f"/api/api/workspaces/{self.visibility.value}", json=desired)
        return {"visibility": self.visibility, "name": self.name, "content_hash": content_hash}

    def update(self, client: httpx.Client, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        if [old_state.visibility, old_state.name] != [self.visibility, self.name]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        desired = self.model_dump(mode="json", exclude_none=True, exclude={"visibility"}, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        if content_hash == getattr(old_state, "content_hash", None):
            return None
        client.request("put", f"/api/api/workspaces/{self.visibility.value}/{self.name}", json=desired)
        return {"visibility": self.visibility, "name": self.name, "content_hash": content_hash}

    @staticmethod
    def delete(client, old_state):
        client.request(
            "delete",
            f"/api/api/workspaces/{old_state.visibility}/{old_state.name}",
        )

    def deps(self) -> Sequence[Resource | Ref]:
        return []


class WorkspaceItemRef(BaseModel, Ref):
    id: str = Field(exclude=True)
    workspace: WorkspaceRef | WorkspaceResource = Field(exclude=True)
    group: str
    name: str

    def attach(self, client):
        viz = self.workspace.visibility.value
        name = self.workspace.name
        try:
            client.request(
                "get",
                f"/api/api/workspaces/{viz}/{name}/items/{self.group}/{self.name}",
            )
        except httpx.HTTPStatusError as ex:
            if ex.response.status_code == 404:
                raise RuntimeError(
                    f"Workspace item {viz}/{name}/items/{self.group}/{self.name} not found"
                )
            else:
                raise ex


@register_resource()
class WorkspaceItemResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    workspace: WorkspaceResource | WorkspaceRef = Field(exclude=True)
    type: str
    group: str
    name: str
    description: str
    content: Dict[Any, Any]
    format: int

    def read(self, client, old_state) -> Dict[str, Any]:
        workspace_path = f"{self.workspace.visibility.value}/{self.workspace.name}"
        remote = client.request(
            "get",
            f"/api/api/workspaces/{workspace_path}/items/{self.group}/{self.name}",
        ).json()
        remote.pop("version")
        remote.pop("links")
        return remote

    def create(self, client: httpx.Client):
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        workspace_path = f"{self.workspace.visibility.value}/{self.workspace.name}"
        client.request(
            "post", f"/api/api/workspaces/{workspace_path}/items", json=desired
        )
        return {
            "visibility": self.workspace.visibility.value,
            "group": self.group,
            "workspace_name": self.workspace.name,
            "name": self.name,
            "content_hash": content_hash,
        }

    def update(self, client: httpx.Client, old_state):
        wksp = self.workspace
        # if we have moved it to a different workspace
        if [old_state.visibility, old_state.workspace_name] != [
            wksp.visibility, wksp.name]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        # rename the item within the workspace
        if [old_state.name, old_state.group] != [self.name, self.group]:
            state = self.create(client)
            self.delete(client, old_state)
            return state
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_json = json.dumps(desired, sort_keys=True)
        content_hash = hashlib.sha256(sorted_json.encode()).hexdigest()
        if content_hash == getattr(old_state, "content_hash", None):
            return None
        workspace_path = f"{self.workspace.visibility.value}/{self.workspace.name}"
        client.request(
            "put",
            f"/api/api/workspaces/{workspace_path}/items/{self.group}/{self.name}",
            json=desired,
        )
        return {
            "visibility": self.workspace.visibility,
            "workspace_name": self.workspace.name,
            "group": self.group,
            "name": self.name,
            "content_hash": content_hash,
        }

    @staticmethod
    def delete(client, old_state):
        workspace_path = f"{old_state.visibility}/{old_state.workspace_name}"
        client.request(
            "delete",
            f"/api/api/workspaces/{workspace_path}/items/{old_state.group}/{old_state.name}",
        )

    def deps(self):
        return [self.workspace]



from __future__ import annotations

from abc import ABC, abstractmethod
from types import SimpleNamespace
from typing import Any, Dict, Optional, Sequence, Union

from httpx import Client as httpxClient
from pydantic import AliasGenerator, ConfigDict, alias_generators

alias_generator = ConfigDict(
    alias_generator=AliasGenerator(
        serialization_alias=lambda field_name: alias_generators.to_camel(field_name)
    )
)

RESOURCE_REGISTRY = {}


def get_resource_type(resource: Resource | Ref) -> str:
    """
    Get the resource name from a resource or ref object.
    """
    resource_type = getattr(resource, "resource_type", None)
    # Use class name if resource_type couldn't be determined
    if not resource_type:
        resource_type = type(resource).__name__

    return resource_type


def register_resource(type_name: str | None = None):
    """
    Decorator to register resource classes.
    Allows lookup of resource classes by their resource_type.
    """

    def decorator(cls):
        # Use class name if resource_type couldn't be determined
        resource_type = type_name if type_name else cls.__name__
        RESOURCE_REGISTRY[resource_type] = {"class": cls}
        cls.resource_type = resource_type
        return cls

    return decorator


def get_resource_class(resource_type: str):
    """
    Get a resource class by its resource_type.
    """
    resource = RESOURCE_REGISTRY.get(resource_type)
    if resource is None:
        raise RuntimeError(f"Resource type '{resource_type}' not found in registry")
    return resource["class"]


class Ref(ABC):
    id: str

    @abstractmethod
    def attach(self, client: httpxClient) -> None:
        pass

    def deps(self):
        return []


class Resource(ABC):
    model_config = alias_generator

    id: str

    @abstractmethod
    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        pass

    @abstractmethod
    def create(self, client: httpxClient) -> Optional[Dict[str, Any]]:
        pass

    @abstractmethod
    def update(self, client: httpxClient, old_state) -> Union[None, Dict[str, Any]]:
        pass

    @staticmethod
    @abstractmethod
    def delete(client: httpxClient, old_state) -> None:
        pass

    @abstractmethod
    def deps(self) -> Sequence["Resource|Ref"]:
        pass


class CamelAlias(ABC):
    model_config = alias_generator



import json
from types import SimpleNamespace
from unittest import mock
from unittest.mock import patch

import httpx
import pytest

import fbnconfig.resource_abc
from fbnconfig import create_client
from fbnconfig.deploy import Action, Deployment, run
from fbnconfig.resource_abc import Resource, get_resource_class, register_resource

TEST_BASE = "https://foo.lusid.com"


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWithNoRemoteResources:
    client = create_client(TEST_BASE, "xyz")

    def test_handles_429_on_get(self, respx_mock):
        # given the server returns a 429 before returning the real response
        respx_mock.get("/api/api/customentities/~deployment").mock(
            side_effect=[
                httpx.Response(429, json={}, headers={"retry-after": "0"}),
                httpx.Response(200, json={"values": []}),
            ]
        )
        # when we run an empty deployment
        deployment = Deployment("whatever", [])
        run(self.client, deployment)
        # then the 429 gets retried and we continue
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert request.url.params._dict == {"filter": ["fields[deployment] eq 'whatever'"]}
        # but no other http requests are made

    def test_run_with_empty_deploy_does_nothing(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        # when we run an empty deployment
        deployment = Deployment("whatever", [])
        actions = run(self.client, deployment)
        # then the log is searched
        request = respx_mock.calls.last.request
        assert request.method == "GET"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert request.url.params._dict == {"filter": ["fields[deployment] eq 'whatever'"]}
        # but no other http requests are made
        # and an empty actions list is returned
        assert actions == []

    def test_run_with_one_resource_creates_it(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # when we run a deployment with one resource
        resource = mock.Mock(spec=Resource)
        resource.deps = mock.Mock(return_value=[])
        resource.id = "23"
        resource.create = mock.Mock(return_value={"some_variable": "x"})
        deployment = Deployment("whatever", [resource])
        actions = run(self.client, deployment)
        # then the create method on the resource is called
        resource.create.assert_called_once_with(self.client)
        # and the create is recorded in the log
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content) == {
            "description": "...",
            "displayName": "23",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "Mock"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "23"},
                {"name": "state", "value": '{"some_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": "deployment",
                    "identifierType": "resource",
                    "identifierValue": "whatever_23",
                }
            ],
        }
        # and the create is in the actions
        assert actions == [Action(id="23", type="Mock", change="create")]

    def test_run_with_one_resource_empty_state(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # when we run a deployment with one resource
        resource = mock.Mock(spec=Resource)
        resource.deps = mock.Mock(return_value=[])
        resource.id = "23"
        # and the new state value is an empty dict
        resource.create = mock.Mock(return_value={})
        deployment = Deployment("whatever", [resource])
        actions = run(self.client, deployment)
        # then the create method on the resource is called
        resource.create.assert_called_once_with(self.client)
        # and the create is recorded in the log
        request = respx_mock.calls.last.request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content) == {
            "description": "...",
            "displayName": "23",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "Mock"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "23"},
                {"name": "state", "value": "{}"},
            ],
            "identifiers": [
                {
                    "identifierScope": "deployment",
                    "identifierType": "resource",
                    "identifierValue": "whatever_23",
                }
            ],
        }
        # and the action is to create
        assert actions == [Action(id="23", type="Mock", change="create")]

    def test_run_with_two_resources(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # and desired resources one and two are independent
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[])
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "y"})
        # when we run a deployment with one and two
        deployment = Deployment("whatever", [resource1, resource2])
        actions = run(self.client, deployment)
        # then the create methods are called on both
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in the order they were
        # added to the deployment
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content)["displayName"] == "two"
        # and they appear in order in the actions
        assert actions == [
            Action(id="one", type="Mock", change="create"),
            Action(id="two", type="Mock", change="create"),
        ]

    def test_run_logs_unique_dependencies(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # and resources "one" and "two" where "two" depends on "one"
        # twice
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1, resource1])  # double dependency
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy
        deployment = Deployment("whatever", [resource1, resource2])
        run(self.client, deployment)
        # then the create method on each of them is called
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in dependency order
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        two_log = json.loads(request.content)
        assert two_log["displayName"] == "two"
        # and resource two has uniquely recorded dependency on resource one
        assert next(
            field["value"] for field in two_log["fields"] if field["name"] == "dependencies"
        ) == ["one"]

    def test_run_with_one_resource_and_dependency(self, respx_mock):
        # given there are no resources in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": []})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # and resources "one" and "two" where "two" depends on "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])
        resource2.id = "two"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy "two" only
        deployment = Deployment("whatever", [resource2])
        actions = run(self.client, deployment)
        # then the create method on each of them is called
        resource1.create.assert_called_once_with(self.client)
        resource2.create.assert_called_once_with(self.client)
        # and the create is recorded in the log for both in dependency order
        request = respx_mock.calls[-2].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        assert json.loads(request.content)["displayName"] == "one"
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        two_log = json.loads(request.content)
        assert two_log["displayName"] == "two"
        # and resource two has recorded a dependency on resource one
        assert next(f["value"] for f in two_log["fields"] if f["name"] == "dependencies") == ["one"]
        # and they appear in order in the actions
        assert actions == [
            Action(id="one", type="Mock", change="create"),
            Action(id="two", type="Mock", change="create"),
        ]

    def test_run_with_deuplicate_resourcei_ids(self, respx_mock):
        # given two resources using the same resource id and a dependency between them
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "duplicate-id"
        resource1.create = mock.Mock(return_value={"some_variable": "x"})
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])
        resource2.id = "duplicate-id"
        resource2.create = mock.Mock(return_value={"some_variable": "x"})
        # when we deploy resource2
        deployment = Deployment("whatever", [resource2])
        # then an error is raised because resource2 depends on resource1
        # but they have the same resource id
        with pytest.raises(RuntimeError) as ex:
            run(self.client, deployment)
        assert "duplicate-id has been used on more than one resource" in str(ex.value)


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeWithExistingResources:
    client = httpx.Client(base_url=TEST_BASE)

    @pytest.fixture
    def existing_resource(self):
        return {
            "description": "...",
            "displayName": "one",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": []},
                {"name": "resource", "value": "one"},
                {"name": "state", "value": '{"some_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": "deployment",
                    "identifierType": "resource",
                    "identifierValue": "whatever_one",
                }
            ],
            "version": {"asAtModified": "2024-02-20T15:39:58.1008350+00:00"},
        }

    @pytest.fixture
    def dependent_resource(self):
        # this resource "two" depends on resource "one"
        return {
            "description": "...",
            "displayName": "two",
            "fields": [
                {"name": "deployment", "value": "whatever"},
                {"name": "resourceType", "value": "FileResource"},
                {"name": "dependencies", "value": ["one"]},
                {"name": "resource", "value": "two"},
                {"name": "state", "value": '{"other_variable": "x"}'},
            ],
            "identifiers": [
                {
                    "identifierScope": "deployment",
                    "identifierType": "resource",
                    "identifierValue": "whatever_two",
                }
            ],
            "version": {"asAtModified": "2024-02-20T15:39:58.1008350+00:00"},
        }

    def test_run_existing_resource_calls_update_no_change(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        # and the desired resource one
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.__name__ = "FileResource"
        # and it is the same as the remote
        resource1.update = mock.Mock(return_value=None)
        # when we deploy "one"
        deployment = Deployment("whatever", [resource1])
        actions = run(self.client, deployment)
        # then the update method is called with the log state
        resource1.update.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
        # and the log is not updated
        # but nochange is in the actions
        assert actions == [Action(id="one", type="Mock", change="nochange")]

    def test_run_existing_resource_calls_update_new_state(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        respx_mock.post("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={})
        )
        # and a desired resource "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.__name__ = "FileResource"
        # and it is different to the remote state (it returns not None)
        resource1.update = mock.Mock(return_value={"some_variable": "abcdefg"})
        # when we deploy "one"
        deployment = Deployment("whatever", [resource1])
        actions = run(self.client, deployment)
        # then the update method is called with the remote log state
        resource1.update.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
        # and the state returned by update is recorded in the log
        request = respx_mock.calls[-1].request
        assert request.method == "POST"
        assert request.url.path == "/api/api/customentities/~deployment"
        content = json.loads(request.content)
        assert content["displayName"] == "one"
        new_state = next(field["value"] for field in content["fields"] if field["name"] == "state")
        assert json.loads(new_state) == {"some_variable": "abcdefg"}
        # and the action is update
        assert actions == [Action(id="one", type="Mock", change="update")]

    def test_run_existing_resource_gets_deleted_when_not_desired(self, respx_mock, existing_resource):
        # given that resource "one" already exists in the log
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource]})
        )
        respx_mock.delete(
            "/api/api/customentities/~deployment/resource/whatever_one?identifierScope=deployment"
        ).mock(return_value=httpx.Response(200, json={}))
        # and "one" is not in the deployment
        deployment = Deployment("whatever", [])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            actions = run(self.client, deployment)
            # then the FileResource.delete function is called (the patched one for this test)
            mock_resource.delete.assert_called_once_with(self.client, SimpleNamespace(some_variable="x"))
            # and the log record is deleted
            request = respx_mock.calls[-1].request
            assert request.method == "DELETE"
            # and the url matches the one mocked above
            # and the action is update
            assert actions == [Action(id="one", type="FileResource", change="remove")]

    def test_run_implicit_resource_does_not_get_deleted(
        self, respx_mock, existing_resource, dependent_resource
    ):
        # given that resource "one" and "two already exists in the log and "two" depends on "one"
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": [existing_resource, dependent_resource]})
        )
        # and desired resources "one" and "two" match what has been deployed
        #    where "two" depends on "one"
        resource1 = mock.Mock(spec=Resource)
        resource1.deps = mock.Mock(return_value=[])
        resource1.id = "one"
        resource1.update = mock.Mock(return_value=None)
        resource2 = mock.Mock(spec=Resource)
        resource2.deps = mock.Mock(return_value=[resource1])  # two depends on one
        resource2.id = "two"
        resource2.update = mock.Mock(return_value=None)
        # and "two" is explicitly in the deployment but "one" is only implicity
        # because two depends on "one"
        deployment = Deployment("whatever", [resource2])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            run(self.client, deployment)
            # then no delete calls are made because we should keep both resources
            assert mock_resource.delete.mock_calls == []

    def test_run_dependent_resources_get_deleted_in_order(
        self, respx_mock, existing_resource, dependent_resource
    ):
        # given resources "one" and "two" where "two"depends on "one"
        respx_mock.get("/api/api/customentities/~deployment").mock(
            return_value=httpx.Response(200, json={"values": [dependent_resource, existing_resource]})
        )
        respx_mock.delete(
            "/api/api/customentities/~deployment/resource/whatever_one?identifierScope=deployment",
            name="delete_one",
        ).mock(return_value=httpx.Response(200, json={}))
        respx_mock.delete(
            "/api/api/customentities/~deployment/resource/whatever_two?identifierScope=deployment",
            name="delete_two",
        ).mock(return_value=httpx.Response(200, json={}))
        # and the deployment is empty
        deployment = Deployment("whatever", [])
        from fbnconfig import drive

        mock_resource = mock.Mock(spec=drive.FileResource)
        mock_resource.delete = mock.Mock(return_value=None)
        with mock.patch(
            "fbnconfig.resource_abc.RESOURCE_REGISTRY", {"FileResource": {"class": mock_resource}}
        ):
            run(self.client, deployment)
            # then the FileResource.delete function is called for both, dependent resource first
            assert mock_resource.delete.mock_calls == [
                mock.call(self.client, SimpleNamespace(other_variable="x")),
                mock.call(self.client, SimpleNamespace(some_variable="x")),
            ]
            # and the log record for two is deleted
            request = respx_mock.calls[-2].request
            assert request.method == "DELETE"
            assert request.url.path == "/api/api/customentities/~deployment/resource/whatever_two"
            # before the log record for one is deleted last
            request = respx_mock.calls[-1].request
            assert request.method == "DELETE"
            assert request.url.path == "/api/api/customentities/~deployment/resource/whatever_one"
            # and the url matches the one mocked above


@pytest.mark.respx(base_url=TEST_BASE)
class DescribeResourceRegistry:
    client = create_client(TEST_BASE, "xyz")

    @pytest.fixture()
    def stub_resource_class(self):
        class StubResource(Resource):
            id: str

            def read(self, client, old_state):
                return None

            def create(self, client):
                return None

            def update(self, client, old_state):
                return None

            @staticmethod
            def delete(client, old_state):
                pass

            def deps(self):
                return []

        return StubResource

    @staticmethod
    def test_resource_registry_returns_expected_registrations(stub_resource_class):
        with patch.dict("fbnconfig.resource_abc.RESOURCE_REGISTRY", {}, clear=True) as registry:
            # GIVEN Two resources are registered
            @register_resource()
            class KlassResource(stub_resource_class):
                pass

            @register_resource(type_name="DifferentName")
            class MyOtherKlassResource(stub_resource_class):
                pass

            # THEN the resource registry is populated with the expected resources
            assert len(registry) == 2
            assert registry == {
                "KlassResource": {"class": KlassResource},
                "DifferentName": {"class": MyOtherKlassResource},
            }

    @staticmethod
    def test_get_resource_class_when_unregistered_resource_fails():
        with pytest.raises(RuntimeError) as ex:
            get_resource_class("UnregisteredResource")

        assert "Resource type 'UnregisteredResource' not found in registry" in str(ex.value)

    @staticmethod
    def test_get_resource_type_defaults_to_name():
        resource1 = mock.Mock(spec=Resource)
        resource1.__name__ = "something"
        type = fbnconfig.resource_abc.get_resource_type(resource1)
        assert type == "Mock"



import json
from hashlib import sha256
from typing import Any, Dict, List, Sequence

from pydantic import BaseModel, Field, computed_field, model_validator

from . import property
from .resource_abc import CamelAlias, Ref, Resource, register_resource


class ResourceId(BaseModel):
    scope: str
    code: str


class AddressKeyList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: str = Field("AddressKeyList", init=False)


class DecimalList(CamelAlias, BaseModel):
    values: list[float]
    reference_list_type: str = Field("DecimalList", init=False)


class FundIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: str = Field("FundIdList", init=False)


class PortfolioGroupIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: str = Field("PortfolioGroupIdList", init=False)


class PortfolioIdList(CamelAlias, BaseModel):
    values: list[ResourceId]
    reference_list_type: str = Field("PortfolioIdList", init=False)


class InstrumentList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: str = Field("InstrumentList", init=False)


class StringList(CamelAlias, BaseModel):
    values: list[str]
    reference_list_type: str = Field("StringList", init=False)


class MetricValue(BaseModel):
    value: float
    unit: str | None


class PropertyListItem(CamelAlias, BaseModel):
    property_definition: property.DefinitionResource | property.DefinitionRef = Field(exclude=True)
    label_value: str | None
    metric_value: MetricValue | None
    label_set_value: Sequence[str] | None
    effective_from: str | None = None
    effective_until: str | None = None

    @computed_field(alias="key")
    def key(self) -> str:
        pd = self.property_definition
        return "/".join([pd.domain.value, pd.scope, pd.code])

    @model_validator(mode="after")
    def validate_one_value_exists(self):
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")
        return self


class PropertyList(CamelAlias, BaseModel):
    values: list[PropertyListItem]
    reference_list_type: str = Field("PropertyList", init=False, exclude=True)


ReferenceListTypes = AddressKeyList | DecimalList | FundIdList | PortfolioGroupIdList | \
    PortfolioIdList | InstrumentList | StringList | PropertyList | PropertyListItem


@register_resource()
class ReferenceListResource(BaseModel, Resource):
    id: str = Field(exclude=True)
    scope: str = Field(exclude=True, init=True)
    code: str = Field(exclude=True, init=True)
    name: str
    description: str | None = None
    tags: list[str] | None = None
    reference_list: ReferenceListTypes

    @computed_field(alias="id")
    def list_id(self) -> dict[str, str]:
        return {"scope": self.scope, "code": self.code}

    def read(self, client, old_state):
        scope, code = old_state.scope, old_state.code
        url = f"/api/api/referencelists/{scope}/{code}"
        entity = client.get(url).json()
        entity.pop("version", None)
        return entity

    def create(self, client) -> Dict[str, Any]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        content_hash = sha256(sorted_desired.encode()).hexdigest()
        client.request("POST", "/api/api/referencelists", json=desired)
        return {"scope": self.scope, "code": self.code, "content_hash": content_hash}

    def update(self, client, old_state):
        if (self.scope, self.code) != (old_state.scope, old_state.code):
            self.delete(client, old_state)
            return self.create(client)
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        sorted_desired = json.dumps(desired, sort_keys=True)
        desired_hash = sha256(sorted_desired.encode()).hexdigest()
        if desired_hash == old_state.content_hash:
            return None
        client.request("POST", "/api/api/referencelists", json=desired)
        return {"scope": self.scope, "code": self.code, "content_hash": desired_hash}

    @staticmethod
    def delete(client, old_state):
        client.request("DELETE", f"/api/api/referencelists/{old_state.scope}/{old_state.code}")

    def deps(self) -> List[Resource | Ref]:
        if isinstance(self.reference_list, PropertyList):
            props = {i.property_definition.id: i.property_definition for i in self.reference_list.values}
            return list(props.values())
        return []



import os
from types import SimpleNamespace

from httpx import HTTPStatusError
from pytest import fixture

import fbnconfig
import tests.integration.configuration as configuration
from tests.integration.generate_test_name import gen

if os.getenv("FBN_ACCESS_TOKEN") is None or os.getenv("LUSID_ENV") is None:
    raise (
        RuntimeError("Both FBN_ACCESS_TOKEN and LUSID_ENV variables need to be set to run these tests")
    )

lusid_env = os.environ["LUSID_ENV"]
token = os.environ["FBN_ACCESS_TOKEN"]
host_vars = {}


@fixture(scope="module")
def setup_deployment():
    deployment_name = gen("configuration")
    print(f"\nRunning for deployment {deployment_name}...")
    yield SimpleNamespace(name=deployment_name)
    # Teardown: Clean up resources (if any) after the test
    print("\nTearing down resources...")
    fbnconfig.deploy(fbnconfig.Deployment(deployment_name, []), lusid_env, token)


def test_teardown(setup_deployment):
    # set it up
    fbnconfig.deploy(configuration.configure(setup_deployment), lusid_env, token)

    # Tear it down
    fbnconfig.deploy(fbnconfig.Deployment(setup_deployment.name, []), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    try:
        client.request("get", "/configuration/api/sets/personal/robTest/rbCode")
    except HTTPStatusError as error:
        # check it was deleted
        assert error.response.status_code == 404


def test_create(setup_deployment):
    fbnconfig.deploy(configuration.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    item = client.request(
        "get", f"/configuration/api/sets/personal/set1-{setup_deployment.name}/rbCode/items/username"
    )
    assert item.status_code == 200


def test_update(setup_deployment):
    fbnconfig.deploy(configuration.configure(setup_deployment), lusid_env, token)
    client = fbnconfig.create_client(lusid_env, token)
    item = client.request(
        "get", f"/configuration/api/sets/personal/set1-{setup_deployment.name}/rbCode/items/username"
    )

    assert item.status_code == 200



import json
from enum import StrEnum
from types import SimpleNamespace
from typing import Any, Dict, List, Optional, Self, Sequence, Union

from httpx import Client as httpxClient
from pydantic import BaseModel, Field, field_serializer, model_validator

from .property import DefinitionRef, DefinitionResource
from .resource_abc import Ref, Resource, register_resource
from .side_definition import SideRef, SideResource


class TransactionTypeAlias(BaseModel):
    type: str
    description: str
    transaction_class: str = Field(serialization_alias="transactionClass")
    transaction_roles: str = Field(serialization_alias="transactionRoles")
    is_default: Optional[bool] = Field(False, serialization_alias="isDefault")


class MovementOption(StrEnum):
    DirectAdjustment = "DirectAdjustment"
    IncludesTradedInterest = "IncludesTradedInterest"


class MovementType(StrEnum):
    Settlement = "Settlement"
    Traded = "Traded"
    StockMovement = "StockMovement"
    FutureCash = "FutureCash"
    Commitment = "Commitment"
    Receivable = "Receivable"
    CashSettlement = "CashSettlement"
    CashForward = "CashForward"
    CashCommitment = "CashCommitment"
    CashReceivable = "CashReceivable"
    Accrual = "Accrual"
    CashAccrual = "CashAccrual"
    ForwardFx = "ForwardFx"
    CashFxForward = "CashFxForward"
    UnsettledCashTypes = "UnsettledCashTypes"
    # Carry movements - Memo transaction to record carry against an instrument.
    Carry = "Carry"  # Carry that will present as an inflow/outflow.
    CarryAsPnl = "CarryAsPnl"  # Carry that will present as a gain or loss.
    VariationMargin = "VariationMargin"
    Capital = "Capital"
    Fee = "Fee"


class CalculationType(StrEnum):
    TaxAmounts = "TaxAmounts"
    NotionalAmount = "Txn:NotionalAmount"
    GrossConsideration = "Txn:GrossConsideration"
    TradeToPortfolioRate = "Txn:TradeToPortfolioRate"
    ExchangeRate = "Txn:ExchangeRate"
    DeriveTotalConsideration = "DeriveTotalConsideration"
    BondInterest = "Txn:BondInterest"


class TransactionTypeCalculation(BaseModel):
    type: CalculationType | str
    side: SideRef | SideResource | None = None
    formula: Optional[str] = None

    @field_serializer("side")
    def serialize_side(self, v: SideRef | SideResource | None):
        if v is not None:
            return v.side
        return None


class MetricValue(BaseModel):
    value: float
    unit: Optional[str] = None


class PerpetualProperty(BaseModel):
    property_key: Union[DefinitionResource, DefinitionRef] = Field(serialization_alias="propertyKey")
    label_value: Optional[str] = Field(default=None, serialization_alias="labelValue")
    metric_value: Optional[MetricValue] = Field(default=None, serialization_alias="metricValue")
    label_set_value: Optional[List[str]] = Field(default=None, serialization_alias="labelSetValue")

    @model_validator(mode="after")
    def validate_one_value_exists(self) -> Self:
        fields = ["label_value", "metric_value", "label_set_value"]
        s = [field for field in fields if getattr(self, field) is not None]
        if len(s) > 1:
            raise KeyError(f"Cannot set {' and '.join(s)}, only one of {' or '.join(fields)} can be set")

        return self


class TransactionTypePropertyMapping(BaseModel):
    property_key: Union[DefinitionResource, DefinitionRef] = Field(serialization_alias="propertyKey")
    map_from: Optional[Union[DefinitionResource, DefinitionRef]] = Field(
        serialization_alias="mapFrom", default=None
    )
    set_to: str = Field(serialization_alias="setTo", default=None)

    @field_serializer("property_key", "map_from", when_used="always")
    def serialize_property_key(self, v: Union[DefinitionRef, DefinitionResource]):
        return f"{v.domain.value}/{v.scope}/{v.code}"


def sort_transaction_type_arrays(obj):  # type: ignore
    if isinstance(obj, dict):
        return {k: sort_transaction_type_arrays(v) for k, v in sorted(obj.items())}
    elif isinstance(obj, list):
        if all(isinstance(i, dict) for i in obj):
            return sorted(
                [sort_transaction_type_arrays(i) for i in obj],
                key=lambda x: json.dumps(x, sort_keys=True),
            )
        else:
            return sorted(obj)
    else:
        return obj


def compare_json_structures(json1, json2) -> bool:
    return sort_transaction_type_arrays(json1) == sort_transaction_type_arrays(json2)


def serialize_properties(values: List[PerpetualProperty]):
    property_dict = {}

    for prop in values:
        value = {}
        if prop.label_value:
            value["labelValue"] = prop.label_value
        elif prop.metric_value:
            value["metricValue"] = prop.metric_value
        elif prop.label_set_value:
            value["labelSetValue"] = prop.label_set_value

        key = f"{prop.property_key.domain.value}/{prop.property_key.scope}/{prop.property_key.code}"
        property_dict[key] = {"key": key, "value": value}

    return property_dict


class TransactionTypeMovement(BaseModel):
    movement_types: MovementType = Field(serialization_alias="movementTypes")
    side: SideRef | SideResource
    direction: int
    properties: Optional[List[PerpetualProperty]] = []
    mappings: Optional[List[TransactionTypePropertyMapping]] = []
    name: Optional[str] = None
    movement_options: Optional[List[MovementOption] | str] = Field(
        default=[], serialization_alias="movementOptions"
    )
    settlement_date_override: Optional[str] = Field(
        default=None, serialization_alias="settlementDateOverride"
    )
    condition: Optional[str] = ""

    @field_serializer("side", when_used="always")
    def serialize_side(self, v: SideRef | SideResource):
        return v.side

    @field_serializer("properties", when_used="always")
    def serialize_properties(self, v: List[PerpetualProperty]):
        return serialize_properties(v)


@register_resource()
class TransactionTypeResource(BaseModel, Resource):
    """
    Represents a transaction type
    """

    id: str = Field(exclude=True, init=True)
    scope: str = Field(exclude=True, init=True)
    source: str = Field(exclude=True, init=True)
    aliases: List[TransactionTypeAlias] = Field(min_length=1)
    movements: List[TransactionTypeMovement] = []
    properties: Optional[List[PerpetualProperty]] = []
    calculations: Optional[List[TransactionTypeCalculation]] = []

    def _get_first_alias(self):
        return sorted([alias.type for alias in self.aliases])[0]

    @field_serializer("properties", when_used="always")
    def serialize_properties(self, v: List[PerpetualProperty]):
        return serialize_properties(v)

    def read(self, client: httpxClient, old_state: SimpleNamespace) -> None | Dict[str, Any]:
        transaction_type = old_state.transaction_type
        source = old_state.source
        scope = old_state.scope
        response = client.get(
            f"/api/api/transactionconfiguration/types/{source}/{transaction_type}",
            params={"scope": scope},
        ).json()
        if "links" in response.keys():
            del response["links"]
        return response

    def create(self, client: httpxClient) -> Optional[Dict[str, Any]]:
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)
        client.put(
            f"/api/api/transactionconfiguration/types/{self.source}/{self._get_first_alias()}",
            params={"scope": self.scope},
            json=desired,
        )
        return {"source": self.source, "scope": self.scope, "transaction_type": self._get_first_alias()}

    def update(self, client: httpxClient, old_state) -> Union[None, Dict[str, Any]]:
        if [old_state.transaction_type, old_state.scope, old_state.source] != [
            self._get_first_alias(),
            self.scope,
            self.source,
        ]:
            self.delete(client, old_state)
            return self.create(client)

        remote = self.read(client, old_state) or {}
        desired = self.model_dump(mode="json", exclude_none=True, by_alias=True)

        if compare_json_structures(remote, desired):
            return None

        return self.create(client)

    @staticmethod
    def delete(client: httpxClient, old_state) -> None:
        transaction_type, scope, source = (old_state.transaction_type, old_state.scope, old_state.source)
        client.delete(
            f"/api/api/transactionconfiguration/types/{source}/{transaction_type}",
            params={"scope": scope},
        )

    def deps(self) -> Sequence[Resource | Ref]:
        def add_to_unique(item):
            if item.id not in seen:
                seen.add(item.id)
                unique.append(item)

        seen = set()
        unique = []

        for movement in self.movements or []:
            if isinstance(movement.side, (SideRef, SideResource)):
                add_to_unique(movement.side)
            for p in movement.properties or []:
                add_to_unique(p.property_key)
            for m in movement.mappings or []:
                add_to_unique(m.property_key)
                if m.map_from:
                    add_to_unique(m.map_from)

        for prop in self.properties or []:
            add_to_unique(prop.property_key)

        for calc in self.calculations or []:
            if calc.side is not None:
                add_to_unique(calc.side)

        return unique
